from setuptools import setup

setup(
    description=
        "Pretty formatter enables pretty formatting using hanging "
        "indents, dataclasses, ellipses, and simple customizability "
        "by registering formatters.",
)
